@@ -12,6 +12,7 @@
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 
+import base64
 import requests
 import os
 import json
@@ -19,6 +20,8 @@ import time
 import urllib.request, urllib.parse, urllib.error
 from .function import Function
 from .workflow import Workflow
+from .storage import Storage,TriggerableBucket
+from .trigger import Trigger
 import logging
 logging.basicConfig()
 #logging.basicConfig(level=logging.DEBUG)
@@ -102,16 +105,21 @@ class MfnClient(object):
 
     @classmethod
     def load_json(cls, config=dict(), filename=None):
-        files = ['../settings.json','~/settings.json','./settings.json']
+        files = ['../settings.json','~/settings.json','~/.mfn/config.json','./settings.json']
         if filename is not None:
             files.append(filename)
         for filename in files:
             """ get the json file """
+            if filename.startswith("~"):
+                filename = os.path.expanduser(filename)
             if os.path.isfile(filename):
                 json_data = {}
-                with open(filename) as json_file:
+                with open(filename,"r") as json_file:
                     try:
-                        json_data = json.load(json_file)
+                        filecontent = json_file.read()
+                        if len(filecontent) == 0:
+                            continue
+                        json_data = json.loads(filecontent)
                         log.info("Using settings file", filename)
                         config.update(json_data)
                     except Exception:
@@ -134,6 +142,10 @@ class MfnClient(object):
         config = MfnClient.load_json()
         config = MfnClient.load_env(config)
         url = mfn_url or config.get('mfn_url')
+        if url is None:
+            print("No valid configuration found")
+            log.warn("Known configuration data:" + str(config))
+            exit(2)
         user = mfn_user or config.get('mfn_user')
         password = mfn_password or config.get('mfn_password')
         name = mfn_name or config.get('mfn_name',None)
@@ -152,6 +164,7 @@ class MfnClient(object):
         self._s = requests.Session()
         if proxies:
             self._s.proxies.update(proxies)
+        self._s.verify=False
         self._s.max_redirects = 10
 
         self.baseurl = str(url)
@@ -163,6 +176,8 @@ class MfnClient(object):
         self.password=password
         self._functions=[]
         self._workflows=[]
+        self._triggers={}
+        self._buckets={}
         try:
             self.login()
         except requests.exceptions.HTTPError as e:
@@ -235,6 +250,7 @@ class MfnClient(object):
         if resp['status'] == 'success':
             self.token = resp['data']['token']
             self.store = self.baseurl+resp['data']['storageEndpoint']
+            self._storage = Storage(self._s, self.token, self.mgmturl)
         else:
             raise Exception("Error logging in at "+self.mgmturl)
 
@@ -517,120 +533,312 @@ class MfnClient(object):
         self.action('deleteWorkflow',{'workflow':{'id':wf.id}})
 
 
-    def keys(self, all_at_once=False):
-        return self.list_keys(all_at_once=all_at_once)
+    # Storage operations
 
-    def keys(self):
+    def _list_stepwise(self, data_type, wid=None):
         start=0
         step=100
         while start >= 0:
-            data_to_send = {}
-            data_to_send["action"] = "performStorageAction"
-            data = {}
-            user = {}
-            user["token"] = self.token
-            data["user"] = user
-            storage = {}
-            storage["action"] = "listKeys"
-            storage["start"] = start
-            storage["count"] = step
-            data["storage"] = storage
-            data_to_send["data"] = data
-            r = self._s.post(self.mgmturl,
-                    params={},
-                    json=data_to_send)
-            r.raise_for_status()
-            if r.json()["status"] == "success":
-                keylist = r.json()["data"]["keylist"]
-                for key in keylist:
-                    yield key
-                if len(keylist) < step:
-                    break
-                start += step
-            else:
-                raise Exception("LISTKEYS failed: " + r.json()["data"]["message"])
+            if data_type == "keys":
+                resultlist = self.list_keys(start, step, wid)
+            elif data_type == "maps":
+                resultlist = self.list_maps(start, step, wid)
+            elif data_type == "sets":
+                resultlist = self.list_sets(start, step, wid)
+            elif data_type == "counters":
+                resultlist = self.list_counters(start, step, wid)
 
-    def list_keys(self, start=0, count=2000):
-        data_to_send = {}
-        data_to_send["action"] = "performStorageAction"
-        data = {}
-        user = {}
-        user["token"] = self.token
-        data["user"] = user
-        storage = {}
-        storage["action"] = "listKeys"
-        storage["start"] = start
-        storage["count"] = count
-        data["storage"] = storage
-        data_to_send["data"] = data
-        r = self._s.post(self.mgmturl,
-                params={},
-                json=data_to_send)
-        r.raise_for_status()
-        if r.json()["status"] != "success":
-            raise Exception("LISTKEYS failed: " + r.json()["data"]["message"])
-
-        return r.json()["data"]["keylist"]
-
-
-    def get(self, key):
-        data_to_send = {}
-        data_to_send["action"] = "performStorageAction"
-        data = {}
-        user = {}
-        user["token"] = self.token
-        data["user"] = user
-        storage = {}
-        storage["action"] = "getdata"
-        storage["key"] = key
-        data["storage"] = storage
-        data_to_send["data"] = data
-        r = self._s.post(self.mgmturl,
-                params={},
-                json=data_to_send)
-        r.raise_for_status()
-        if r.json()["status"] == "success":
-            return r.json()["data"]["value"]
+            for result in resultlist:
+                yield result
+            if len(resultlist) < step:
+                break
+            start += step
         else:
-            raise Exception("GET failed: " + r.json()["data"]["message"])
+            raise Exception("LIST" + data_type.upper() + " failed: " + r.json()["data"]["message"])
+
+    # kv operations
+    def get(self, key, wid=None):
+        return self._storage.get(key, wid)
+
+    def put(self, key, value, wid=None):
+        self._storage.put(key, value, wid)
+
+    def delete(self, key, wid=None):
+        self._storage.delete(key, wid)
+
+    def list_keys(self, start=0, count=2000, wid=None):
+        return self._storage.list_keys(start, count, wid)
+
+    def keys(self, wid=None):
+        self._list_stepwise("keys", wid)
+
+    # map operations
+    def create_map(self, mapname, wid=None):
+        self._storage.create_map(mapname, wid)
+
+    def put_map_entry(self, mapname, key, value, wid=None):
+        self._storage.put_map_entry(mapname, key, value, wid)
+
+    def get_map_entry(self, mapname, key, wid=None):
+        return self._storage.get_map_entry(mapname, key, wid)
+
+    def delete_map_entry(self, mapname, key, wid=None):
+        self._storage.delete_map_entry(mapname, key, wid)
+
+    def retrieve_map(self, mapname, wid=None):
+        return self._storage.retrieve_map(mapname, wid)
+
+    def contains_map_key(self, mapname, key, wid=None):
+        return self._storage.contains_map_key(mapname, key, wid)
+
+    def get_map_keys(self, mapname, wid=None):
+        return self._storage.get_map_keys(mapname, wid)
+
+    def clear_map(self, mapname, wid=None):
+        self._storage.clear_map(mapname, wid)
+
+    def delete_map(self, mapname, wid=None):
+        self._storage.delete_map(mapname)
+
+    def list_maps(self, start=0, count=2000, wid=None):
+        return self._storage.list_maps(start, count, wid)
+
+    def maps(self, wid=None):
+        self._list_stepwise("maps", wid)
+
+    # set operations
+    def create_set(self, setname, wid=None):
+        self._storage.create_set(setname, wid)
+
+    def add_set_entry(self, setname, item, wid=None):
+        self._storage.add_set_entry(setname, item, wid)
+
+    def remove_set_entry(self, setname, item, wid=None):
+        self._storage.remove_set_entry(setname, item, wid)
+
+    def contains_set_item(self, setname, item, wid=None):
+        return self._storage.contains_set_item(setname, item, wid)
+
+    def retrieve_set(self, setname, wid=None):
+        return self._storage.retrieve_set(setname, wid)
+
+    def clear_set(self, setname, wid=None):
+        self._storage.clear_set(setname, wid)
+
+    def delete_set(self, setname, wid=None):
+        self._storage.delete_set(setname)
+
+    def list_sets(self, start=0, count=2000, wid=None):
+        return self._storage.list_sets(start, count, wid)
+
+    def sets(self, wid=None):
+        self._list_stepwise("sets", wid)
+
+    # counter operations
+    def create_counter(self, countername, countervalue, wid=None):
+        self._storage.create_counter(countername, countervalue, wid)
+
+    def get_counter(self, countername, wid=None):
+        return self._storage.get_counter(countername, wid)
+
+    def increment_counter(self, countername, increment, wid=None):
+        self._storage.increment_counter(countername, increment, wid)
+
+    def decrement_counter(self, countername, decrement, wid=None):
+        self._storage.decrement_counter(countername, decrement, wid)
+
+    def delete_counter(self, countername, wid=None):
+        self._storage.delete_counter(countername, wid)
+
+    def list_counters(self, start=0, count=2000, wid=None):
+        return self._storage.list_counters(start, count, wid)
+
+    def counters(self, wid=None):
+        self._list_stepwise("counters", wid)
+
+    ### Triggers
+    @property
+    def triggers(self):
+        data = self.action('getTriggerDetails',{'trigger_names':[]})
+        unseen = list(self._triggers.keys())
+        for tname,tdetails in data['trigger_details'].items():
+            if tname in self._triggers:
+                self._triggers[tname].update(tdetails)
+                unseen.remove(tname)
+            else:
+                self._triggers[tname] = Trigger(self,tname,tdetails)
+        for tname in unseen:
+            del self._triggers[tname]
+        return self._triggers
+
+    def add_trigger(self,name,config):
+        """ add a trigger
+
+        requires a name and a trigger configuration, raises an error if it already exists, creates a new trigger if it doesn't exist and returns the new Trigger object
+
+        :name: name of the trigger
+        :config: trigger configuration (trigger_info)
+        """
+        if name in self.triggers.keys():
+            return self.triggers[name]
+        try:
+            data = self.action('addTrigger',{'trigger_name':name,'trigger_info':config})
+            ts = self.get_triggers([name])
+            if len(ts) == 0:
+                raise Exception("Trigger '"+name+"' was just created but not found")
+            self.triggers[name] = ts[0]
+        except Exception as e:
+            raise e
+
+    def get_triggers(self,names):
+        """ fetch trigger details for specific triggers
+
+        requires a non-empty list of trigger names to fetch, returns an array of triggers found for the given names
+        :names: list of names to fetch
+        """
+        try:
+            data = self.action('getTriggerDetails',{'trigger_names':names})
+            for tname, tdata in data["trigger_details"].items():
+                self._triggers[tname] = Trigger(self, tname, tdata)
+        except Exception as e:
+            raise e
+
+    def find_trigger(self,name):
+        res = []
+        for n,t in self.triggers.items():
+            if n == name:
+                return t
+            elif n.startswith(name):
+                res.append(t)
+        if len(res) == 0:
+            raise Exception("Could not find trigger matching "+name)
+        else:
+            if(len(res) > 1):
+                log.warning("Found multiple triggers for search word "+name+", returning first result")
+            return res[0]
+
+    def bind_trigger(self,trigger_name,workflow_name):
+        """ binds a workflow to a trigger
+
+        Bind a workflow to a trigger (aka 'associate' a workflow with a trigger)
+        raises Exception if not successful 
+        :trigger_name: name of the trigger (immutable)
+        :workflow_name: name of the workflow as which it has been bound
+        """
+        try:
+            self.action('addTriggerForWorkflow',{'trigger_name':trigger_name,'workflow_name':workflow_name})
+        except requests.exceptions.HTTPError as e:
+            e.strerror += "while trying to bind workflow '"+workflow_name+"' to trigger '"+trigger_name+"'"
+            raise e
+    
+    def unbind_trigger(self,trigger_name,workflow_name):
+        """ unbinds a workflow from a trigger
+        Unbind a workflow from a trigger (aka 'disassociate' a workflow from a trigger)
+        raises Exception if not successful 
+        :trigger_name: name of the trigger (immutable)
+        :workflow_name: name of the workflow as which it has been bound
+        """
+        try:
+            self.action('deleteTriggerForWorkflow',{'trigger_name':trigger_name,'workflow_name':workflow_name})
+        except requests.exceptions.HTTPError as e:
+            e.strerror += "while trying to unbind workflow '"+workflow_name+"' from trigger '"+trigger_name+"'"
+            raise e
+
+    def delete_trigger(self,trigger_name):
+        """ deletes a trigger
+
+        Delete a trigger (automatically diassociates all bound workflows)
+        :trigger_name: name of the trigger (immutable)
+        """
+        try:
+            self.action('deleteTrigger',{'trigger_name':trigger_name})
+        except requests.exceptions.HTTPError as e:
+            e.strerror += "while trying to delete trigger '"+trigger_name+"'"
+            raise e
 
 
-    def put(self, key, value):
-        data_to_send = {}
-        data_to_send["action"] = "performStorageAction"
-        data = {}
-        user = {}
-        user["token"] = self.token
-        data["user"] = user
-        storage = {}
-        storage["action"] = "putdata"
-        storage["key"] = key
-        storage["value"] = value
-        data["storage"] = storage
-        data_to_send["data"] = data
-        r = self._s.post(self.mgmturl,
-                params={},
-                json=data_to_send)
-        r.raise_for_status()
-        if r.json()["status"] != "success":
-            raise Exception("PUT failed: " + r.json()["data"]["message"])
+    ### Buckets
+    @property
+    def buckets(self):
+        data = self.action('getTriggerableBuckets')
+        unseen = list(self._buckets.keys())
+        for bname,bworkflows in data['buckets'].items():
+            bmetadatalist = data['buckets_details'].get(bname,[])
+            if bname in self._buckets:
+                self._buckets[bname]._update(bworkflows, bmetadatalist)
+                unseen.remove(bname)
+            else:
+                self._buckets[bname] = TriggerableBucket(self,bname,bworkflows,bmetadatalist)
+        for bname in unseen:
+            del self._buckets[name]
+        return self._buckets
 
+    def find_bucket(self,name):
+        res = []
+        for n,b in self.buckets.items():
+            if n == name:
+                return b
+            elif n.startswith(name):
+                res.append(b)
+        if len(res) == 0:
+            raise Exception("Could not find bucket matching "+name)
+        else:
+            if(len(res) > 1):
+                log.warning("Found multiple buckets for search word "+name+", returning first result")
+            return res[0]
 
-    def delete(self, key):
-        data_to_send = {}
-        data_to_send["action"] = "performStorageAction"
-        data = {}
-        user = {}
-        user["token"] = self.token
-        data["user"] = user
-        storage = {}
-        storage["action"] = "deletedata"
-        storage["key"] = key
-        data["storage"] = storage
-        data_to_send["data"] = data
-        r = self._s.post(self.mgmturl,
-                params={},
-                json=data_to_send)
-        r.raise_for_status()
-        if r.json()["status"] != "success":
-            raise Exception("DELETE failed: " + r.json()["data"]["message"])
+    def add_bucket(self,bname):
+        """ add a bucket
+
+        requires a name, creates a new bucket if it doesn't exist and returns the new Trigger object
+
+        :name: name of the trigger
+        :config: trigger configuration (trigger_info)
+        """
+        if bname in self._buckets:
+            return self._buckets[bname] 
+        try:
+            data = self.action('addTriggerableBucket',{'bucketname':bname})
+            self._buckets[bname] = TriggerableBucket(self, bname)
+            return self._buckets[bname] 
+        except Exception as e:
+            raise e
+
+    def bind_bucket(self,bucket_name,workflow_name):
+        """ binds a workflow to a bucket
+
+        Bind a workflow to a bucket (aka 'associate' a workflow with a bucket)
+        raises Exception if not successful 
+        :bucket_name: name of the bucket
+        :workflow_name: name of the workflow as which it has been bound
+        """
+        try:
+            self.action('addStorageTriggerForWorkflow',{'bucketname':bucket_name,'workflowname':workflow_name})
+        except requests.exceptions.HTTPError as e:
+            e.strerror += "while trying to bind workflow '"+workflow_name+"' to bucket '"+bucket_name+"'"
+            raise e
+    
+    def unbind_bucket(self,bucket_name,workflow_name):
+        """ unbinds a workflow from a bucket
+        Unbind a workflow from a bucket (aka 'disassociate' a workflow from a bucket)
+        raises Exception if not successful 
+        :bucket_name: name of the bucket
+        :workflow_name: name of the workflow as which it has been bound
+        """
+        try:
+            self.action('deleteStorageTriggerForWorkflow',{'tablename':bucket_name,'workflowname':workflow_name})
+        except requests.exceptions.HTTPError as e:
+            e.strerror += "while trying to unbind workflow '"+workflow_name+"' from bucket '"+bucket_name+"'"
+            raise e
+
+    def delete_bucket(self,bucket_name):
+        """ deletes a bucket
+
+        Delete a bucket (automatically diassociates all bound workflows)
+        :bucket_name: name of the bucket
+        """
+        try:
+            self.action('deleteTriggerableBucket',{'bucketname':bucket_name})
+        except requests.exceptions.HTTPError as e:
+            e.strerror += "while trying to delete bucket '"+bucket_name+"'"
+            raise e
